#!/usr/bin/env python

"""
Common calendar client utilities.

Copyright (C) 2014, 2015 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from datetime import datetime
from imiptools.data import Object, get_address, get_uri, get_window_end, \
                           is_new_object, make_freebusy, to_part, \
                           uri_dict, uri_items, uri_values
from imiptools.dates import format_datetime, get_default_timezone, \
                            get_recurrence_start_point, get_timestamp, \
                            to_timezone
from imiptools.period import can_schedule, remove_period, \
                             remove_additional_periods, remove_affected_period, \
                             update_freebusy
from imiptools.profile import Preferences
import imip_store

def update_attendees(obj, attendees, removed):

    """
    Update the attendees in 'obj' with the given 'attendees' and 'removed'
    attendee lists. A list is returned containing the attendees whose
    attendance should be cancelled.
    """

    to_cancel = []

    existing_attendees = uri_values(obj.get_values("ATTENDEE") or [])
    added = set(attendees).difference(existing_attendees)

    if added or removed:
        attendees = uri_items(obj.get_items("ATTENDEE") or [])
        sequence = obj.get_value("SEQUENCE")

        if removed:
            remaining = []

            for attendee, attendee_attr in attendees:
                if attendee in removed:

                    # Without a sequence number, assume that the event has not
                    # been published and that attendees can be silently removed.

                    if sequence is not None:
                        to_cancel.append((attendee, attendee_attr))
                else:
                    remaining.append((attendee, attendee_attr))

            attendees = remaining

        if added:
            for attendee in added:
                attendee = attendee.strip()
                if attendee:
                    attendees.append((get_uri(attendee), {"PARTSTAT" : "NEEDS-ACTION", "RSVP" : "TRUE"}))

        obj["ATTENDEE"] = attendees

    return to_cancel

class Client:

    "Common handler and manager methods."

    default_window_size = 100

    def __init__(self, user, messenger=None, store=None, publisher=None):
        self.user = user
        self.messenger = messenger
        self.store = store or imip_store.FileStore()

        try:
            self.publisher = publisher or imip_store.FilePublisher()
        except OSError:
            self.publisher = None

        self.preferences = None

    def get_preferences(self):
        if not self.preferences and self.user:
            self.preferences = Preferences(self.user)
        return self.preferences

    def get_tzid(self):
        prefs = self.get_preferences()
        return prefs and prefs.get("TZID") or get_default_timezone()

    def get_window_size(self):
        prefs = self.get_preferences()
        try:
            return prefs and int(prefs.get("window_size")) or self.default_window_size
        except (TypeError, ValueError):
            return self.default_window_size

    def get_window_end(self):
        return get_window_end(self.get_tzid(), self.get_window_size())

    def is_sharing(self):
        prefs = self.get_preferences()
        return prefs and prefs.get("freebusy_sharing") == "share" or False

    def is_bundling(self):
        prefs = self.get_preferences()
        return prefs and prefs.get("freebusy_bundling") == "always" or False

    def is_notifying(self):
        prefs = self.get_preferences()
        return prefs and prefs.get("freebusy_messages") == "notify" or False

    # Common operations on calendar data.

    def is_participating(self, user, as_organiser=False):

        """
        Return whether, subject to the 'user' indicating an identity and the
        'as_organiser' status of that identity, the user concerned is actually
        participating in the current object event.
        """

        attr = self.get_attendance(user)
        return as_organiser or not attr or attr.get("PARTSTAT") != "DECLINED"

    def get_overriding_transparency(self, user, as_organiser=False):

        """
        Return the overriding transparency to be associated with the free/busy
        records for an event, subject to the 'user' indicating an identity and
        the 'as_organiser' status of that identity.

        Where an identity is only an organiser and not attending, "ORG" is
        returned. Otherwise, no overriding transparency is defined and None is
        returned.
        """

        attr = self.get_attendance(user)
        return as_organiser and not (attr and attr.get("PARTSTAT")) and "ORG" or None

    def update_participation(self, obj, partstat=None):

        """
        Update the participation in 'obj' of the user with the given 'partstat'.
        """

        attendee_attr = uri_dict(obj.get_value_map("ATTENDEE")).get(self.user)
        if not attendee_attr:
            return None
        if partstat:
            attendee_attr["PARTSTAT"] = partstat
        if attendee_attr.has_key("RSVP"):
            del attendee_attr["RSVP"]
        self.update_sender(attendee_attr)
        return attendee_attr

    def update_sender(self, attr):

        "Update the SENT-BY attribute of the 'attr' sender metadata."

        if self.messenger and self.messenger.sender != get_address(self.user):
            attr["SENT-BY"] = get_uri(self.messenger.sender)

    def get_periods(self, obj):

        """
        Return periods for the given 'obj'. Interpretation of periods can depend
        on the time zone, which is obtained for the current user.
        """

        return obj.get_periods(self.get_tzid(), self.get_window_end())

    # Store operations.

    def get_stored_object(self, uid, recurrenceid):

        """
        Return the stored object for the current user, with the given 'uid' and
        'recurrenceid'.
        """

        fragment = self.store.get_event(self.user, uid, recurrenceid)
        return fragment and Object(fragment)

    # Free/busy operations.

    def get_freebusy_part(self, freebusy=None):

        """
        Return a message part containing free/busy information for the user,
        either specified as 'freebusy' or obtained from the store directly.
        """

        if self.is_sharing() and self.is_bundling():

            # Invent a unique identifier.

            utcnow = get_timestamp()
            uid = "imip-agent-%s-%s" % (utcnow, get_address(self.user))

            freebusy = freebusy or self.store.get_freebusy(self.user)

            user_attr = {}
            self.update_sender(user_attr)
            return to_part("PUBLISH", [make_freebusy(freebusy, uid, self.user, user_attr)])

        return None

    def update_freebusy(self, freebusy, periods, transp, uid, recurrenceid, summary, organiser):

        """
        Update the 'freebusy' collection with the given 'periods', indicating a
        'transp' status, explicit 'uid' and 'recurrenceid' to indicate either a
        recurrence or the parent event. The 'summary' and 'organiser' must also
        be provided.
        """

        update_freebusy(freebusy, periods, transp, self.uid, recurrenceid, summary, organiser)

class ClientForObject(Client):

    "A client maintaining a specific object."

    def __init__(self, obj, user, messenger=None, store=None, publisher=None):
        Client.__init__(self, user, messenger, store, publisher)
        self.set_object(obj)

    def set_object(self, obj):

        "Set the current object to 'obj', obtaining metadata details."

        self.obj = obj
        self.uid = obj and self.obj.get_uid()
        self.recurrenceid = obj and self.obj.get_recurrenceid()
        self.sequence = obj and self.obj.get_value("SEQUENCE")
        self.dtstamp = obj and self.obj.get_value("DTSTAMP")

    # Object update methods.

    def update_dtstamp(self):

        "Update the DTSTAMP in the current object."

        dtstamp = self.obj.get_utc_datetime("DTSTAMP")
        utcnow = to_timezone(datetime.utcnow(), "UTC")
        self.obj["DTSTAMP"] = [(format_datetime(dtstamp > utcnow and dtstamp or utcnow), {})]

    def set_sequence(self, increment=False):

        "Update the SEQUENCE in the current object."

        sequence = self.obj.get_value("SEQUENCE") or "0"
        self.obj["SEQUENCE"] = [(str(int(sequence) + (increment and 1 or 0)), {})]

    def merge_attendance(self, attendees):

        """
        Merge attendance from the current object's 'attendees' into the version
        stored for the current user.
        """

        obj = self.get_stored_object_version()

        if not obj or not self.have_new_object(obj):
            return False

        # Get attendee details in a usable form.

        attendee_map = uri_dict(obj.get_value_map("ATTENDEE"))

        for attendee, attendee_attr in attendees.items():

            # Update attendance in the loaded object.

            attendee_map[attendee] = attendee_attr

        # Set the new details and store the object.

        obj["ATTENDEE"] = attendee_map.items()

        # Set the complete event if not an additional occurrence.

        event = obj.to_node()
        self.store.set_event(self.user, self.uid, self.recurrenceid, event)

        return True

    # Object-related tests.

    def get_attendance(self, user=None):

        """
        Return the attendance attributes for 'user', or the current user if
        'user' is not specified.
        """

        attendees = uri_dict(self.obj.get_value_map("ATTENDEE"))
        return attendees.get(user or self.user) or {}

    def is_attendee(self, identity, obj=None):

        """
        Return whether 'identity' is an attendee in the current object, or in
        'obj' if specified.
        """

        return identity in uri_values((obj or self.obj).get_values("ATTENDEE"))

    def can_schedule(self, freebusy, periods):

        """
        Indicate whether within 'freebusy' the given 'periods' can be scheduled.
        """

        return can_schedule(freebusy, periods, self.uid, self.recurrenceid)

    def have_new_object(self, obj=None):

        """
        Return whether the current object is new to the current user (or if the
        given 'obj' is new).
        """

        obj = obj or self.get_stored_object_version()

        # If found, compare SEQUENCE and potentially DTSTAMP.

        if obj:
            sequence = obj.get_value("SEQUENCE")
            dtstamp = obj.get_value("DTSTAMP")

            # If the request refers to an older version of the object, ignore
            # it.

            return is_new_object(sequence, self.sequence, dtstamp, self.dtstamp,
                self.is_partstat_updated(obj))

        return True

    def is_partstat_updated(self, obj):

        """
        Return whether the participant status has been updated in the current
        object in comparison to the given 'obj'.

        NOTE: Some clients like Claws Mail erase time information from DTSTAMP
        NOTE: and make it invalid. Thus, such attendance information may also be
        NOTE: incorporated into any new object assessment.
        """

        old_attendees = uri_dict(obj.get_value_map("ATTENDEE"))
        new_attendees = uri_dict(self.obj.get_value_map("ATTENDEE"))

        for attendee, attr in old_attendees.items():
            old_partstat = attr.get("PARTSTAT")
            new_attr = new_attendees.get(attendee)
            new_partstat = new_attr and new_attr.get("PARTSTAT")

            if old_partstat == "NEEDS-ACTION" and new_partstat and \
               new_partstat != old_partstat:

                return True

        return False

    # Object retrieval.

    def get_stored_object_version(self):

        """
        Return the stored object to which the current object refers for the
        current user.
        """

        return self.get_stored_object(self.uid, self.recurrenceid)

    def get_definitive_object(self, from_organiser):

        """
        Return an object considered definitive for the current transaction,
        using 'from_organiser' to select the current transaction's object if
        true, or selecting a stored object if false.
        """

        return from_organiser and self.obj or self.get_stored_object_version()

    def get_parent_object(self):

        """
        Return the parent object to which the current object refers for the
        current user.
        """

        return self.recurrenceid and self.get_stored_object(self.uid, None) or None

    # Convenience methods for modifying free/busy collections.

    def get_recurrence_start_point(self, recurrenceid):

        "Get 'recurrenceid' in a form suitable for matching free/busy entries."

        tzid = self.obj.get_tzid() or self.get_tzid()
        return get_recurrence_start_point(recurrenceid, tzid)

    def remove_from_freebusy(self, freebusy):

        "Remove this event from the given 'freebusy' collection."

        if not remove_period(freebusy, self.uid, self.recurrenceid) and self.recurrenceid:
            remove_affected_period(freebusy, self.uid, self.get_recurrence_start_point(self.recurrenceid))

    def remove_freebusy_for_recurrences(self, freebusy, recurrenceids=None):

        """
        Remove from 'freebusy' any original recurrence from parent free/busy
        details for the current object, if the current object is a specific
        additional recurrence. Otherwise, remove all additional recurrence
        information corresponding to 'recurrenceids', or if omitted, all
        recurrences.
        """

        if self.recurrenceid:
            recurrenceid = self.get_recurrence_start_point(self.recurrenceid)
            remove_affected_period(freebusy, self.uid, recurrenceid)
        else:
            # Remove obsolete recurrence periods.

            remove_additional_periods(freebusy, self.uid, recurrenceids)

            # Remove original periods affected by additional recurrences.

            if recurrenceids:
                for recurrenceid in recurrenceids:
                    recurrenceid = self.get_recurrence_start_point(recurrenceid)
                    remove_affected_period(freebusy, self.uid, recurrenceid)

    def update_freebusy(self, freebusy, user, for_organiser):

        """
        Update the 'freebusy' collection for this event with the periods and
        transparency associated with the current object, subject to the 'user'
        identity and the attendance details provided for them, indicating
        whether the update is 'for_organiser' or not.
        """

        # Obtain the stored object if the current object is not issued by the
        # organiser. Attendees do not have the opportunity to redefine the
        # periods.

        obj = self.get_definitive_object(for_organiser)
        if not obj:
            return

        # Obtain the affected periods.

        periods = self.get_periods(obj)

        # Define an overriding transparency, the indicated event transparency,
        # or the default transparency for the free/busy entry.

        transp = self.get_overriding_transparency(user, for_organiser) or \
                 obj.get_value("TRANSP") or \
                 "OPAQUE"

        # Perform the low-level update.

        Client.update_freebusy(self, freebusy, periods, transp,
            self.uid, self.recurrenceid,
            obj.get_value("SUMMARY"),
            obj.get_value("ORGANIZER"))

    def update_freebusy_for_participant(self, freebusy, user, for_organiser=False,
                                        updating_other=False):

        """
        Update the 'freebusy' collection using the given 'periods', involving
        the given 'user', indicating whether the update is 'for_organiser' or
        not, and whether it is 'updating_other' (meaning another user's
        details).
        """

        # Record in the free/busy details unless a non-participating attendee.
        # Use any attendee information for an organiser, not the organiser's own
        # attributes.

        if self.is_participating(user, for_organiser and not updating_other):
            self.update_freebusy(freebusy, user, for_organiser)
        else:
            self.remove_from_freebusy(freebusy)

    # Convenience methods for updating stored free/busy information received
    # from other users.

    def update_freebusy_from_participant(self, user, for_organiser):

        """
        For the current user, record the free/busy information for another
        'user', indicating whether the update is 'for_organiser' or not, thus
        maintaining a separate record of their free/busy details.
        """

        # A user does not store free/busy information for themself as another
        # party.

        if user == self.user:
            return

        freebusy = self.store.get_freebusy_for_other(self.user, user)
        self.update_freebusy_for_participant(freebusy, user, for_organiser, True)

        # Tidy up any obsolete recurrences.

        self.remove_freebusy_for_recurrences(freebusy, self.store.get_recurrences(self.user, self.uid))
        self.store.set_freebusy_for_other(self.user, freebusy, user)

    def update_freebusy_from_organiser(self, organiser):

        "For the current user, record free/busy information from 'organiser'."

        self.update_freebusy_from_participant(organiser, True)

    def update_freebusy_from_attendees(self, attendees):

        "For the current user, record free/busy information from 'attendees'."

        for attendee in attendees.keys():
            self.update_freebusy_from_participant(attendee, False)

# vim: tabstop=4 expandtab shiftwidth=4
